/* Copyright 2012, Unpublished Work of Technologic Systems
 * All Rights Reserved.
 *
 * THIS WORK IS AN UNPUBLISHED WORK AND CONTAINS CONFIDENTIAL,
 * PROPRIETARY AND TRADE SECRET INFORMATION OF TECHNOLOGIC SYSTEMS.
 * ACCESS TO THIS WORK IS RESTRICTED TO (I) TECHNOLOGIC SYSTEMS 
 * EMPLOYEES WHO HAVE A NEED TO KNOW TO PERFORM TASKS WITHIN THE SCOPE
 * OF THEIR ASSIGNMENTS AND (II) ENTITIES OTHER THAN TECHNOLOGIC
 * SYSTEMS WHO HAVE ENTERED INTO APPROPRIATE LICENSE AGREEMENTS.  NO
 * PART OF THIS WORK MAY BE USED, PRACTICED, PERFORMED, COPIED, 
 * DISTRIBUTED, REVISED, MODIFIED, TRANSLATED, ABRIDGED, CONDENSED, 
 * EXPANDED, COLLECTED, COMPILED, LINKED, RECAST, TRANSFORMED, ADAPTED
 * IN ANY FORM OR BY ANY MEANS, MANUAL, MECHANICAL, CHEMICAL, 
 * ELECTRICAL, ELECTRONIC, OPTICAL, BIOLOGICAL, OR OTHERWISE WITHOUT
 * THE PRIOR WRITTEN PERMISSION AND CONSENT OF TECHNOLOGIC SYSTEMS.
 * ANY USE OR EXPLOITATION OF THIS WORK WITHOUT THE PRIOR WRITTEN
 * CONSENT OF TECHNOLOGIC SYSTEMS COULD SUBJECT THE PERPETRATOR TO
 * CRIMINAL AND CIVIL LIABILITY.
 */

/* As the name implies, this module bridges PCIe to wishbone.  It connects to
 * one PCIe lane as an endpoint.  On the wishbone side there is an "IO" 
 * wishbone master, which is used for simple memory read and write requests
 * on the PCIe bus, and eight DMA masters, each of which is capable of
 * streaming data in either direction.  Outside this module, 32 bit wishbone
 * slaves should typically be connected to the IO master (wbm_*) and one or the
 * individual DMA masters.
 *
 * It can be debated whether "DMA" is the correct term here.  I was going to
 * call it bus mastering, but everyone seems to want to say DMA, and the DMA
 * modules are in fact accessing memory directly, albeit over the PCIe link.
 *
 * Each DMA channel has 64 bits of control input and 14 bits of COUNT output
 * that is used to track the progress of transfers.  These signals can be
 * connected externally to this module to wishbone register blocks controlled
 * by wbm_*, or to other logic that wants to control transfers.
 *
 * The register map of an individual DMA channel control inputs is as follows:
 * Bits    Name    Function
 * ----    ----    --------
 * 63:50   SIZE    Number of DWords to transfer
 * 49      INCR    Increment endpoint address
 * 48      SLOW    Throttle based on incoming IRQ
 * 47:34   EADR    Endpoint start address
 * 33      DIR     Transfer direction
 * 32      ENA     Enable transfer
 * 31:2    RADR    Root start address
 * 1       IRQE    Enable IRQ to indicate complete transfer
 * 0       ROLL    Rollover: repeat transfer indefinitely
 *
 * The IO wishbone master correctly handles 16 bit or 8 bit requests, but the
 * DMA can only transfer 32 bit quantities.  External logic is provided
 * for using this DMA controller with 16 bit and 8 bit wishbone cores.
 *
 * EADR is the FPGA start address.  It can address 64KB.  While more than 64KB
 * of address space is provided in the FPGA, each individual DMA channel is
 * restricted to a 64KB region.  If INCR is zero, the DMA transfers will 
 * read or write all words to EADR.
 *
 * DIR dictates the direction.  If set to 0, data is moved from the endpoint to
 * the root, replacing a CPU read. If set to 1, data is moved from the root to
 * the endpoint, replacing a CPU write. COUNT indicates an estimate of how
 * many DWwords are remaining to be transfered.
 *
 * After transfering SIZE DWords, the transfer will stop, unless ROLL is set to
 * one, in which case it will automatically start over at EADR and RADR and
 * repeat the transfer.  If using ROLL, the region starting at RADR is acting
 * as a ring buffer, and software needs to monitor COUNT and act accordingly.
 *
 * ROLL probably does not make sense without also using SLOW.  If using SLOW,
 * the slave device provides an IRQ input to the DMA block which indicates
 * whether the slave is ready for a read or write.  Transfers only occur when
 * the slave is ready.
 *
 * If ROLL is not used, then data transfer stops when the transfer is complete.
 * In order to start a new transfer, software must deassert and then assert
 * ENA.
 *
 * If IRQE is set, the IRQ output will be asserted when a transfer is complete.
 * To clear the IRQ, deassert ENA.
 *
 * If DIR is changed while a transfer is in progress, it will probably crash
 * the whole board.  Confirm that COUNT is zero before modifying DIR.
 */

module pcie_wb_bridge (
  wb_rst_i,
  wb_clk_i,

  pcie_rx_p_i,
  pcie_rx_m_i,
  pcie_tx_p_o,
  pcie_tx_m_o,
  pcie_100mhz_i,

  wbm_stb_o,
  wbm_cyc_o,
  wbm_adr_o,
  wbm_sel_o,
  wbm_we_o,
  wbm_dat_o,
  wbm_dat_i,
  wbm_ack_i,

  ch0_stb_o,
  ch0_cyc_o,
  ch0_adr_o,
  ch0_sel_o,
  ch0_we_o,
  ch0_dat_o,
  ch0_dat_i,
  ch0_ack_i,
  dma_ctl0_i,
  count0_o,

  ch1_stb_o,
  ch1_cyc_o,
  ch1_adr_o,
  ch1_sel_o,
  ch1_we_o,
  ch1_dat_o,
  ch1_dat_i,
  ch1_ack_i,
  dma_ctl1_i,
  count1_o,

  ch2_stb_o,
  ch2_cyc_o,
  ch2_adr_o,
  ch2_sel_o,
  ch2_we_o,
  ch2_dat_o,
  ch2_dat_i,
  ch2_ack_i,
  dma_ctl2_i,
  count2_o,

  ch3_stb_o,
  ch3_cyc_o,
  ch3_adr_o,
  ch3_sel_o,
  ch3_we_o,
  ch3_dat_o,
  ch3_dat_i,
  ch3_ack_i,
  dma_ctl3_i,
  count3_o,

  ch4_stb_o,
  ch4_cyc_o,
  ch4_adr_o,
  ch4_sel_o,
  ch4_we_o,
  ch4_dat_o,
  ch4_dat_i,
  ch4_ack_i,
  dma_ctl4_i,
  count4_o,

  ch5_stb_o,
  ch5_cyc_o,
  ch5_adr_o,
  ch5_sel_o,
  ch5_we_o,
  ch5_dat_o,
  ch5_dat_i,
  ch5_ack_i,
  dma_ctl5_i,
  count5_o,

  ch6_stb_o,
  ch6_cyc_o,
  ch6_adr_o,
  ch6_sel_o,
  ch6_we_o,
  ch6_dat_o,
  ch6_dat_i,
  ch6_ack_i,
  dma_ctl6_i,
  count6_o,

  ch7_stb_o,
  ch7_cyc_o,
  ch7_adr_o,
  ch7_sel_o,
  ch7_we_o,
  ch7_dat_o,
  ch7_dat_i,
  ch7_ack_i,
  dma_ctl7_i,
  count7_o,

  dma_irq_i,
  irq_i,

  dma_disable_i
);

input wb_rst_i;
input wb_clk_i;

input pcie_rx_p_i;
input pcie_rx_m_i;
output pcie_tx_p_o;
output pcie_tx_m_o;
input pcie_100mhz_i;

// main wishbone master
output wbm_stb_o;
output wbm_cyc_o;
output [31:0] wbm_adr_o;
output [3:0] wbm_sel_o;
output wbm_we_o;
output [31:0] wbm_dat_o;
input [31:0] wbm_dat_i;
input wbm_ack_i;

// 8 DMA channel masters
output ch0_stb_o;
output ch0_cyc_o;
output [31:0] ch0_adr_o;
output [3:0] ch0_sel_o;
output ch0_we_o;
output [31:0] ch0_dat_o;
input [31:0] ch0_dat_i;
input ch0_ack_i;
input [63:0] dma_ctl0_i;
output [13:0] count0_o;

output ch1_stb_o;
output ch1_cyc_o;
output [31:0] ch1_adr_o;
output [3:0] ch1_sel_o;
output ch1_we_o;
output [31:0] ch1_dat_o;
input [31:0] ch1_dat_i;
input ch1_ack_i;
input [63:0] dma_ctl1_i;
output [13:0] count1_o;

output ch2_stb_o;
output ch2_cyc_o;
output [31:0] ch2_adr_o;
output [3:0] ch2_sel_o;
output ch2_we_o;
output [31:0] ch2_dat_o;
input [31:0] ch2_dat_i;
input ch2_ack_i;
input [63:0] dma_ctl2_i;
output [13:0] count2_o;

output ch3_stb_o;
output ch3_cyc_o;
output [31:0] ch3_adr_o;
output [3:0] ch3_sel_o;
output ch3_we_o;
output [31:0] ch3_dat_o;
input [31:0] ch3_dat_i;
input ch3_ack_i;
input [63:0] dma_ctl3_i;
output [13:0] count3_o;

output ch4_stb_o;
output ch4_cyc_o;
output [31:0] ch4_adr_o;
output [3:0] ch4_sel_o;
output ch4_we_o;
output [31:0] ch4_dat_o;
input [31:0] ch4_dat_i;
input ch4_ack_i;
input [63:0] dma_ctl4_i;
output [13:0] count4_o;

output ch5_stb_o;
output ch5_cyc_o;
output [31:0] ch5_adr_o;
output [3:0] ch5_sel_o;
output ch5_we_o;
output [31:0] ch5_dat_o;
input [31:0] ch5_dat_i;
input ch5_ack_i;
input [63:0] dma_ctl5_i;
output [13:0] count5_o;

output ch6_stb_o;
output ch6_cyc_o;
output [31:0] ch6_adr_o;
output [3:0] ch6_sel_o;
output ch6_we_o;
output [31:0] ch6_dat_o;
input [31:0] ch6_dat_i;
input ch6_ack_i;
input [63:0] dma_ctl6_i;
output [13:0] count6_o;

output ch7_stb_o;
output ch7_cyc_o;
output [31:0] ch7_adr_o;
output [3:0] ch7_sel_o;
output ch7_we_o;
output [31:0] ch7_dat_o;
input [31:0] ch7_dat_i;
input ch7_ack_i;
input [63:0] dma_ctl7_i;
output [13:0] count7_o;

input [7:0] dma_irq_i;
input [7:0] irq_i;

input [7:0] dma_disable_i;

endmodule

