/* Copyright 2009, Unpublished Work of Technologic Systems
 * All Rights Reserved.
 *
 * THIS WORK IS AN UNPUBLISHED WORK AND CONTAINS CONFIDENTIAL,
 * PROPRIETARY AND TRADE SECRET INFORMATION OF TECHNOLOGIC SYSTEMS.
 * ACCESS TO THIS WORK IS RESTRICTED TO (I) TECHNOLOGIC SYSTEMS 
 * EMPLOYEES WHO HAVE A NEED TO KNOW TO PERFORM TASKS WITHIN THE SCOPE
 * OF THEIR ASSIGNMENTS AND (II) ENTITIES OTHER THAN TECHNOLOGIC
 * SYSTEMS WHO HAVE ENTERED INTO APPROPRIATE LICENSE AGREEMENTS.  NO
 * PART OF THIS WORK MAY BE USED, PRACTICED, PERFORMED, COPIED, 
 * DISTRIBUTED, REVISED, MODIFIED, TRANSLATED, ABRIDGED, CONDENSED, 
 * EXPANDED, COLLECTED, COMPILED, LINKED, RECAST, TRANSFORMED, ADAPTED
 * IN ANY FORM OR BY ANY MEANS, MANUAL, MECHANICAL, CHEMICAL, 
 * ELECTRICAL, ELECTRONIC, OPTICAL, BIOLOGICAL, OR OTHERWISE WITHOUT
 * THE PRIOR WRITTEN PERMISSION AND CONSENT OF TECHNOLOGIC SYSTEMS.
 * ANY USE OR EXPLOITATION OF THIS WORK WITHOUT THE PRIOR WRITTEN
 * CONSENT OF TECHNOLOGIC SYSTEMS COULD SUBJECT THE PERPETRATOR TO
 * CRIMINAL AND CIVIL LIABILITY.
 */

/* Register map summary:
 * 0x00: model ID
 * 0x04: LEDs and other control bits
 * 0x08: interrupt mask
 * 0x0c: WDT feed register
 * 0x10: MUXBUS config register
 * 0x14: usec counter
 * 0x20: ICAP controls
 * 0x40 to 0x4c: DIO output values
 * 0x50 to 0x5c: DIO direction registers
 * 0x60 to 0x6c: DIO input values
 *
 * I could have fit all these registers in a 0-0x3f space, but it would have
 * been ugly, less usable, and less flexible when features come along in the
 * future.  We are not low on LUTs.
 *
 */

module syscon(
  wb_clk_i,
  wb_rst_i,
  clk_100mhz_i,
  rst_100mhz_i,

  wb_cyc_i,
  wb_stb_i,
  wb_we_i,
  wb_adr_i,
  wb_dat_i,
  wb_sel_i,
  wb_dat_o,
  wb_ack_o,

  dio_i,
  dio_oe_o,
  dio_o,
  control_reg_o,
  bus_config_o,
  irq_mask_o,

  mode2_i,
  reboot_o
);

input clk_100mhz_i, rst_100mhz_i;

input wb_clk_i, wb_rst_i, wb_cyc_i, wb_stb_i, wb_we_i;
input [31:0] wb_adr_i;
input [31:0] wb_dat_i;
input [3:0] wb_sel_i;
output [31:0] wb_dat_o;
output wb_ack_o;

input [107:0] dio_i;
output [107:0] dio_oe_o;
output [107:0] dio_o;

output [31:0] control_reg_o;
output [15:0] bus_config_o;
output [7:0] irq_mask_o;

input mode2_i;
output reboot_o;

localparam model = 16'h4300;
localparam rev = 12'h107;

reg icap_clk, icap_en, icap_we;
reg [15:0] icap_dat;
wire [15:0] icap_swap = {
  icap_dat[8],
  icap_dat[9],
  icap_dat[10],
  icap_dat[11],
  icap_dat[12],
  icap_dat[13],
  icap_dat[14],
  icap_dat[15],
  icap_dat[0],
  icap_dat[1],
  icap_dat[2],
  icap_dat[3],
  icap_dat[4],
  icap_dat[5],
  icap_dat[6],
  icap_dat[7]
};

ICAP_SPARTAN6
  //#( .DEVICE_ID(32'h04024093)) // Specifies the pre-programmed Device ID value
icapcore (
  //.BUSY(BUSY),   // 1-bit output: Busy/Ready output
  //.O(O),         // 16-bit output: Configuartion data output bus
  .CE(icap_en),       // 1-bit input: Active-Low ICAP Enable input
  .CLK(icap_clk),     // 1-bit input: Clock input
  .I(icap_swap),         // 16-bit input: Configuration data input bus
  .WRITE(icap_we)  // 1-bit input: Read/Write control input
);

reg [107:0] dio_oe, dio;
assign dio_oe_o = dio_oe;
assign dio_o = dio;
reg [31:0] control_reg;
assign control_reg_o = control_reg;
reg [15:0] bus_config;
assign bus_config_o = bus_config;
reg [7:0] irq_mask;
assign irq_mask_o = irq_mask;

reg [1:0] wdt_setting;
reg reboot;
reg [31:0] wdt_count;
assign reboot_o = reboot || control_reg[31] || (!dio_i[9] && control_reg[30]);
always @(*) case(wdt_setting)
  2'd0: reboot = wdt_count[26];  // 0.67s // should be [25] maybe
  2'd1: reboot = wdt_count[28];  // 2.68s
  2'd2: reboot = wdt_count[30];  // 10.74s
  2'd3: reboot = 1'b0;
endcase
always @(posedge wb_clk_i or posedge wb_rst_i) begin
  if (wb_rst_i) begin
    wdt_setting <= 2'd3;
    wdt_count <= 32'd0;
  end else begin
    wdt_count <= wdt_count + 1'b1;
    if (wdt_setting == 2'd3) wdt_count <= 32'd0;
    if (wb_cyc_i && wb_stb_i && wb_we_i && (wb_adr_i[7:0] == 8'hc)) begin
      wdt_setting <= wb_dat_i[1:0];
      wdt_count <= 32'd0;
    end
  end
end

/* 100MHz is used for the counter so that wb_clk can be changed to any 
 * frequency without changing the counter logic.
 */
reg [7:0] usec_count;
reg usec0;
always @(posedge clk_100mhz_i or posedge rst_100mhz_i) begin
  if (rst_100mhz_i) begin
    usec_count <= 8'd0;
    usec0 <= 1'b0;
  end else begin
    usec_count <= usec_count + 1'b1;
    if (usec_count == 8'd99) begin
      usec_count <= 8'd0;
      usec0 <= ~usec0;
    end
  end
end

reg [31:0] usec;
reg [1:0] usec_q;
always @(posedge wb_clk_i or posedge wb_rst_i) begin
  if (wb_rst_i) begin
    usec <= 32'd0;
    usec_q <= 2'd0;
  end else begin
    usec_q[1] <= usec_q[0];
    usec_q[0] <= usec0;
    if ((usec_q[1] == usec_q[0]) && (usec_q[0] != usec[0]))
      usec <= usec + 1'b1;
  end
end

always @(posedge wb_clk_i or posedge wb_rst_i) begin
  if (wb_rst_i) begin
    dio_oe <= 108'd0;
    dio <= 108'd0;
    control_reg <= 32'h00800003; // 2 LEDs default to on; LM32 defaults in rst
    bus_config <= 16'd0;
    irq_mask <= 8'hff; // mask out all by default
    icap_clk <= 1'b1;
    icap_dat <= 16'd0;
    icap_en <= 1'b1;
    icap_we <= 1'b1;
  end else begin
    if (wb_cyc_i && wb_stb_i && wb_we_i) case (wb_adr_i[6:0])
      7'h04: control_reg <= wb_dat_i;
      7'h08: irq_mask <= wb_dat_i[7:0];
      7'h10: bus_config <= wb_dat_i[15:0];
      7'h20: {icap_we, icap_en, icap_clk, icap_dat} <= wb_dat_i[18:0];
      7'h40: dio[31:0] <= wb_dat_i;
      7'h44: dio[63:32] <= wb_dat_i;
      7'h48: dio[95:64] <= wb_dat_i;
      7'h4c: dio[107:96] <= wb_dat_i[11:0];
      7'h50: dio_oe[31:0] <= wb_dat_i;
      7'h54: dio_oe[63:32] <= wb_dat_i;
      7'h58: dio_oe[95:64] <= wb_dat_i;
      7'h5c: dio_oe[107:96] <= wb_dat_i[11:0];
    endcase
  end
end

reg wb_ack;
assign wb_ack_o = wb_ack;
reg [31:0] wb_dat;
assign wb_dat_o = wb_dat;
always @(posedge wb_clk_i or posedge wb_rst_i) begin
  if (wb_rst_i) begin
    wb_dat <= 32'd0;
    wb_ack <= 1'b0;
  end else begin
    wb_ack <= wb_cyc_i && wb_stb_i && !wb_ack;
    case (wb_adr_i[6:0]) 
      7'h00: wb_dat <= {3'd0, mode2_i, rev, model};
      7'h04: wb_dat <= control_reg;
      7'h08: wb_dat <= {24'd0, irq_mask};
      7'h10: wb_dat <= {16'd0, bus_config};
      7'h14: wb_dat <= usec;
      7'h20: wb_dat <= {13'd0, icap_we, icap_en, icap_clk, icap_dat};
      7'h40: wb_dat <= dio[31:0];
      7'h44: wb_dat <= dio[63:32];
      7'h48: wb_dat <= dio[95:64];
      7'h4c: wb_dat <= {20'd0, dio[107:96]};
      7'h50: wb_dat <= dio_oe[31:0];
      7'h54: wb_dat <= dio_oe[63:32];
      7'h58: wb_dat <= dio_oe[95:64];
      7'h5c: wb_dat <= {20'd0, dio_oe[107:96]};
      7'h60: wb_dat <= dio_i[31:0];
      7'h64: wb_dat <= dio_i[63:32];
      7'h68: wb_dat <= dio_i[95:64];
      7'h6c: wb_dat <= {20'd0, dio_i[107:96]};
    endcase
  end
end

endmodule

